from vcScript import *
from vcHelpers.Selection import *
import vcVector

global status, hold_until
status = "Grasping"
hold_until = None
finalized = False
trigger_node = None
trigger_joint = None
grasping_signal = None
default_detection_volume = vcVector.new(50,50,50)

def product_transition(product,arrived):
  global status, hold_until

  if arrived == True: # GRASP
    status = "HoldingUntilMinTimeDone" 
  else: # RELEASE
    status = "Grasping" 
    hold_until = None
  evaluateCondition()
  

def OnRun():
  global status
  if control_mode_prop is not None and control_mode_prop.Value != "Emulation":
    return
  
  while True:
    # condition(lambda: status == "Grasping" or status == "HoldingUntilMinTimeDone")
    condition(lambda: status != "ReadyToRelease")
    if status == "HoldingUntilMinTimeDone":
      delay(delay_grasp.Value)
      condition(lambda: status == "HoldingUntilMinTimeDone")
      status = "ReadyToRelease"
    elif status == "Grasping":
      grasping_signal.signal(True)
      delay(interval_p.Value)
    elif status == "WaitUntilRegraspDelayDone":
      delay(delay_release.Value)
      status = "Grasping"

#Register to the OnNodeTransform event
def OnStart():
  global trigger_node, trigger_joint, status
  #if control_mode_prop is not None and control_mode_prop.Value != "Emulation":
  #  return
  trigger_node = get_trigger_node()
  # if finger_node is not None and enabled_prop.Value:
  if trigger_node is not None:
    trigger_joint = trigger_node.Dof
    trigger_node.OnNodeTransform = handler
  status = "Grasping"

#Unregister to the OnNodeTransform event
def OnReset():
  global trigger_node
  if trigger_node is not None:
    trigger_node.OnNodeTransform = None

#return the node that should trigger the grasping
def get_trigger_node():
  if release_joint_p.Value is None:
    return None
  split_name = release_joint_p.Value.split(":")
  split_name = split_name[0].strip()
  trigger_node = comp.findNode(split_name)
  return trigger_node

#Main event handler for triggering the grasp signal
def handler(arg):
  global status, last_peak_value, last_value, last_direction
  if not sim.IsRunning:
    return
  
  delta = trigger_delta_prop.Value

  if release_mode_p.Value == "increasing to decreasing":
    if trigger_joint.VALUE < last_value: # value decreasing
      if last_direction == "increasing" and abs(trigger_joint.VALUE-last_peak_value) > delta:
        last_direction = "decreasing"
        if status == "ReadyToRelease":
          grasping_signal.signal(False)
          status = "WaitUntilRegraspDelayDone"
          evaluateCondition()
      else:
        last_direction = "increasing"
        last_peak_value = trigger_joint.VALUE

  elif release_mode_p.Value == "decreasing to increasing":
    if trigger_joint.VALUE > last_value: # value increasing
      if last_direction == "decreasing" and (last_peak_value-trigger_joint.VALUE) > delta:
        last_direction = "increasing"
        if status == "ReadyToRelease":
          grasping_signal.signal(False)
          status = "WaitUntilRegraspDelayDone"
          evaluateCondition()
      else:
        last_direction = "decreasing"
        last_peak_value = trigger_joint.VALUE

  last_value = trigger_joint.VALUE

#Update the trigger joint property step values when the comp nodes are changed (new node might have been added)
def update_trigger_options(type_constant, node1, node2):
  if type_constant == VC_NODE_REMOVE_NODE and node1 == comp: #Component deleted
    return
  elif type_constant != VC_NODE_REMOVE_NODE: #Nodes were modified
    set_trigger_servo(node1)
  else: #Node other than the component was removed
    set_trigger_servo(comp)

#Executed when the component is loaded
def OnFinalize():
  global grasping_signal, finalized, container
  if finalized:
    return
  finalized = True
  grasping_signal = grasping_node.getBehaviour(grasping_node.Name + " Signal")
  container = grasping_node.getBehaviour(grasping_node.Name + " Container__HIDE__")
  if grasping_signal is None:
    print "Error. Failed to find the grasping signal"
  container.OnTransition = product_transition
  init_trigger_joint()

#Initializes the TriggerJoint value and step constraints
def init_trigger_joint():
  parent_node = grasping_node.Parent
  trigger_node = set_trigger_servo(parent_node)
  if trigger_node is None:
    print "Failed to set the TriggerJoint"

#Initialize Trigger Servo to the other child of the grasping node
def set_trigger_servo(selected_node):
  children = selected_node.Children
  all_children = getGivenComponentsNodes(comp)
  #Combine the two lists without duplicates
  children = list(children)
  children.extend(x for x in all_children if x not in children)
  if None in children:
    return None
  nodes_and_dofs, trigger_node = build_trigger_step_values(children)
  if not nodes_and_dofs:
    print("No valid servos found for Trigger Servo")
    return None
  release_joint_p.StepValues = nodes_and_dofs
  if release_joint_p.Value is None:
    release_joint_p.Value = release_joint_p.StepValues[0]
  return trigger_node

#Return a list of Node name & Dof name, and the first trigger node to be the default value
def build_trigger_step_values(children):
  trigger_node = None
  nodes_and_dofs = []
  VALID_JOINT_TYPES = [VC_JOINTTYPE_TRANSLATIONAL, VC_JOINTTYPE_ROTATIONAL, VC_JOINTTYPE_CUSTOM]
  for c in children:
    if c.Dof is None or c.Dof.JointLinkType not in VALID_JOINT_TYPES:
      continue
    if trigger_node == None:
      trigger_node = c
    combined_names = "{} : {}".format(c.Name, c.Dof.Name)
    nodes_and_dofs.append(combined_names)
  return nodes_and_dofs, trigger_node

def force_prod_locs(p):
  if force_locs_p.Value:
    force_pick_p.IsVisible = True
    force_place_p.IsVisible = True
    print "Type in the axis that you want to force in ForceProductOffset: X, Y, and/or Z"
  else:
    force_pick_p.IsVisible = False
    force_place_p.IsVisible = False

#Validate that the given property exists, and return the property
def confirmProperty(name,defaultvalue,type=VC_REAL,constraints=VC_PROPERTY_DEFAULT):
     prop = comp.getProperty(name)
     if type == VC_VECTOR:
        if prop == None:
          prop = comp.createProperty(type,name,constraints)
          if defaultvalue:
            prop.Value = vcVector.new()
          else:
            prop.Value = default_detection_volume
     else:
        if prop == None:
          prop = comp.createProperty(type,name,constraints)
          if constraints == VC_PROPERTY_STEP:
            prop.StepValues = DEFAULT_TRIGGER_JOINT_VALS
          prop.Value = defaultvalue
     return prop

global last_peak_value, last_value, last_direction

comp = getComponent()
sim = getSimulation()
grasping_node = getNode()

DEFAULT_EPSILON = 0.5
DEFAULT_INTERVAL = 0.05
DEFAULT_MIN_PRODUCT_HOLD = 0.1 
DEFAULT_MIN_REGRASP_DELAY = 0.3
DEFAULT_RELEASE_TRIGGER_JOINT = 'J3 : J3' 
DEFAULT_RELEASE_TRIGGER_MODE = 'decreasing to increasing'
DEFAULT_TRIGGER_JOINT_VALS = []
DEFAULT_FORCE_PICK_PLACE_LOCATIONS = ""
DEFAULT_FORCE_PICK = 0.0
DEFAULT_FORCE_PLACE = 0.0
# DEFAULT_RELEASE_TRIGGER_MODE_VALS = []

last_peak_value = None
last_value = None
last_direction = None
interval_p = confirmProperty(grasping_node.Name + "::GraspDetectionInterval", DEFAULT_INTERVAL, VC_REAL)
delay_grasp = confirmProperty(grasping_node.Name + "::DelayAfterGrasp", DEFAULT_MIN_PRODUCT_HOLD, VC_REAL)
release_joint_p = confirmProperty(grasping_node.Name + "::ReleaseTriggerJoint", DEFAULT_RELEASE_TRIGGER_JOINT, VC_STRING, VC_PROPERTY_STEP)
release_mode_p = confirmProperty(grasping_node.Name + "::ReleaseTriggerMode", DEFAULT_RELEASE_TRIGGER_MODE, VC_STRING, VC_PROPERTY_STEP)
release_mode_p.StepValues = ["decreasing to increasing","increasing to decreasing"]
trigger_delta_prop = confirmProperty(grasping_node.Name + "::ReleaseTriggerDelta", DEFAULT_EPSILON, VC_REAL)
delay_release = confirmProperty(grasping_node.Name + "::DelayAfterRelease", DEFAULT_MIN_REGRASP_DELAY, VC_REAL)

force_locs_p = confirmProperty(grasping_node.Name + "::ForceProductOffset", DEFAULT_FORCE_PICK_PLACE_LOCATIONS, VC_STRING)
# force_locs_p.Value = False
force_locs_p.OnChanged = force_prod_locs
force_pick_p = confirmProperty(grasping_node.Name + "::PickOffset", DEFAULT_FORCE_PICK, VC_VECTOR)
force_place_p = confirmProperty(grasping_node.Name + "::PlaceOffset", DEFAULT_FORCE_PLACE, VC_VECTOR)
force_prod_locs(force_locs_p)

# release_joint_p.StepValues = ['Cable : Cable', 'J1 : J1', 'J2 : J2', 'J3 : J3', 'J4 : J4']
control_mode_prop = comp.getProperty("ControlMode")

det_volume = comp.getProperty(grasping_node.Name + "::DetectionVolumeSize")
det_volume.Value = vcVector.new(5,5,5)
det_volume.IsVisible = True
gravity_dir = comp.getProperty(grasping_node.Name + "::GravityDirection")
gravity_dir.Value = 5
comp.OnNodeConfigurationChange = update_trigger_options
OnFinalize()
