from vcCommand import *
from vcHelpers.Selection import *
import vcVector

#Original release v2.0

app = getApplication()
cmd = getCommand()

GRASP_NODE_NAME = "Grasp"

nodeProp = cmd.createProperty(VC_STRING, "Grasping Node", VC_PROPERTY_STEP)
mode = cmd.createProperty(VC_STRING, "Mode", VC_PROPERTY_STEP)
mode.StepValues = ["Default","Servo Trigger","Collision Trigger"]
mode.Value = ""

# MODE - Add Servo Trigger
servoProp = cmd.createProperty(VC_STRING, "Trigger Servo", VC_PROPERTY_STEP)
servoProp.IsVisible = False
triggerPointProp = cmd.createProperty(VC_REAL, "Trigger Point")
triggerPointProp.IsVisible = False

# MODE - Collision Trigger
det_interval = cmd.createProperty(VC_REAL, "GraspDetectionInterval")
det_interval.Value = 0.1
det_interval.IsVisible = False
delay_grasp = cmd.createProperty(VC_REAL, "DelayAfterGrasp")
delay_grasp.Value = 0.3
delay_grasp.IsVisible = False
servoProp = cmd.createProperty(VC_STRING, "ReleaseTriggerJoint", VC_PROPERTY_STEP)
servoProp.IsVisible = False
rel_mode = cmd.createProperty(VC_STRING, "ReleaseTriggerMode", VC_PROPERTY_STEP)
rel_mode.StepValues = ["increasing to decreasing","decreasing to increasing"]
rel_mode.Value = "increasing to decreasing"
rel_mode.IsVisible = False

# MODE - Common
deltaProp = cmd.createProperty(VC_REAL, "ReleaseTriggerDelta")
deltaProp.IsVisible = False

# MODE - Collision Trigger 2
delay_release = cmd.createProperty(VC_REAL, "DelayAfterRelease")
delay_release.Value = 0.3
delay_release.IsVisible = False
force_locs_p = cmd.createProperty(VC_STRING,"ForceProductOffset")
force_locs_p.Value = ""
force_pick_p = cmd.createProperty(VC_VECTOR,"PickOffset")
force_pick_p.IsVisible = False
force_pick_p.Value = vcVector.new()
force_place_p = cmd.createProperty(VC_VECTOR,"PlaceOffset")
force_place_p.IsVisible = False
force_place_p.Value = vcVector.new()

applyProp = cmd.createProperty(VC_BUTTON, "Apply")

#
def first_state():
  global applyProp, comp, mode
  mode.Value = "Default"
  context = app.CurrentContext
  if context.Id != VC_CONTEXT_AUTHOR:
    warning('This command can be called from modeling context only')
    return
  comp = context.ActiveComponent
  if not comp:
    warning('Please select one component')
    return
  selected_node = context.ActiveNode
  set_default_node(selected_node)
  set_default_trigger(selected_node)
  
  executeInActionPanel()
  applyProp.OnChanged = applyWizard

#
def force_prod_locs(p):
  if force_locs_p.Value:
    force_pick_p.IsVisible = True
    force_place_p.IsVisible = True
    print "Type in the axis that you want to force in ForceProductOffset: X, Y, and/or Z"
  else:
    force_pick_p.IsVisible = False
    force_place_p.IsVisible = False
  
def set_default_node(selected_node):
  selected_node_index = -1
  nodes = getGivenComponentsNodes(comp)
  nodeProp.StepValues = [node.Name for node in nodes]
  if selected_node is not None and selected_node != selected_node.Component:
    for i, name in enumerate(nodeProp.StepValues):
      if name == selected_node.Name:
        selected_node_index = i
        break
  nodeProp.Value = nodeProp.StepValues[selected_node_index]

#Choose DOF from available servos. Preferably the other child from the common parent node

def set_default_trigger(selected_node):
  default_servo = set_trigger_servo(selected_node)
  set_trigger_point(default_servo)


#Initialize Trigger Servo to the other child of the grasping node
def set_trigger_servo(selected_node):
  trigger_servo = None
  children = selected_node.Children
  all_children = getGivenComponentsNodes(comp)
  #Combine the two lists without duplicates
  children = list(children)
  children.extend(x for x in all_children if x not in children)
  #List of Node name, Dof name
  nodes_and_dofs = []
  VALID_JOINT_TYPES = [VC_JOINTTYPE_TRANSLATIONAL, VC_JOINTTYPE_ROTATIONAL, VC_JOINTTYPE_CUSTOM]
  for c in children:
    if c.Dof is not None and c.Dof.JointLinkType in VALID_JOINT_TYPES:
      if trigger_servo == None:
        trigger_servo = c
      combined_names = "{} : {}".format(c.Name, c.Dof.Name)
      #print c.Name, c.Dof.Name, c.Dof.JointLinkType, c.Dof.JointServoType
      nodes_and_dofs.append(combined_names)
  if not nodes_and_dofs:
    warning("No valid servos found for Trigger Servo")
    return None
  servoProp.StepValues = nodes_and_dofs #[c.Name for c in children]
  servoProp.Value = servoProp.StepValues[-1]
  return trigger_servo


#Automatically use min and max joint limits (1/2 of the min max as default rigger point and 10% of that as the delta)
def set_trigger_point(default_servo):
  DELTA_MIN = 0.5
  DELTA_MAX = 10.0
  triggerPointProp.Value = 0.0
  deltaProp.Value = DELTA_MIN
  if default_servo is None:
    return
  try:
    min_limit = float(default_servo.Dof.MinLimit)
    max_limit = float(default_servo.Dof.MaxLimit)
  except ValueError:
    min_limit = None
    max_limit = None
  if min_limit is not None and max_limit is not None:
    range = max_limit - min_limit
    default_trigger_point = min_limit + range/2
    triggerPointProp.Value = default_trigger_point
    default_delta = default_trigger_point / 10.0
    default_delta = clamp(default_delta, DELTA_MIN, DELTA_MAX)
    deltaProp.Value = default_delta

#
def applyWizard(*args):
  sel = app.SelectionManager.getSelection(VC_SELECTION_COMPONENT)
  if not len(sel) == 1:
    print "Select one component"
    return
  comp = sel[0]
  # comp = app.CurrentContext.ActiveComponent
  createProperties(comp)
  parentNode = comp.findNode(nodeProp.Value)
  node = createNode(parentNode)
  createFeatures(node)
  createBehaviors(node)
  print cmd.Name, "added grasping functionalities to", node.Name
  app.render()

#
def set_trigger_parameters(script_content):
  if mode.Value == "Servo Trigger":
    default_string = "DEFAULT_EPSILON = 0.5"
    new_string = "DEFAULT_EPSILON = {}".format(deltaProp.Value)
    script_content = script_content.replace(default_string, new_string)
    default_string = "DEFAULT_TRIGGER_POINT = 5.0"
    new_string = "DEFAULT_TRIGGER_POINT = {}".format(triggerPointProp.Value)
    script_content = script_content.replace(default_string, new_string)
    default_string = "DEFAULT_TRIGGER_JOINT = ''"
    new_string = "DEFAULT_TRIGGER_JOINT = '{}'".format(servoProp.Value)
    script_content = script_content.replace(default_string, new_string)

  elif mode.Value == "Collision Trigger":
    default_string = "DEFAULT_INTERVAL = 0.1"
    new_string = "DEFAULT_INTERVAL = {}".format(det_interval.Value)
    script_content = script_content.replace(default_string, new_string)
    default_string = "DEFAULT_MIN_PRODUCT_HOLD = 0.3"
    new_string = "DEFAULT_MIN_PRODUCT_HOLD = {}".format(delay_grasp.Value)
    script_content = script_content.replace(default_string, new_string)
    default_string = "DEFAULT_delay_release = 0.3"
    new_string = "DEFAULT_delay_release = {}".format(delay_release.Value)
    script_content = script_content.replace(default_string, new_string)
    default_string = "DEFAULT_RELEASE_TRIGGER_JOINT = ''"
    new_string = "DEFAULT_RELEASE_TRIGGER_JOINT = '{}'".format(servoProp.Value)
    script_content = script_content.replace(default_string, new_string)
    default_string = "DEFAULT_RELEASE_TRIGGER_MODE = ''"
    new_string = "DEFAULT_RELEASE_TRIGGER_MODE = '{}'".format(rel_mode.Value)
    script_content = script_content.replace(default_string, new_string)
    default_string = 'DEFAULT_FORCE_PICK_PLACE_LOCATIONS = ""'
    new_string = 'DEFAULT_FORCE_PICK_PLACE_LOCATIONS = "{}"'.format(force_locs_p.Value)
    script_content = script_content.replace(default_string, new_string)
    default_string = "DEFAULT_FORCE_PICK = 0.0"
    new_string = "DEFAULT_FORCE_PICK = {}".format([force_pick_p.Value.X, force_pick_p.Value.Y, force_pick_p.Value.Z])
    script_content = script_content.replace(default_string, new_string)
    default_string = "DEFAULT_FORCE_PLACE = 0.0"
    new_string = "DEFAULT_FORCE_PLACE = {}".format([force_place_p.Value.X, force_place_p.Value.Y, force_place_p.Value.Z])
    script_content = script_content.replace(default_string, new_string)
    
    #print "rel_mode.StepValues: ",rel_mode.StepValues
    # new_string = "DEFAULT_RELEASE_TRIGGER_MODE_VALS = '{}'".format(rel_mode.StepValues)
    # script_content = script_content.replace(default_string, new_string)
  
  return script_content

def getScript(scriptName):
  import os

  cmdUri = getCommandPath().replace(r"file:///", '')
  head, tail = os.path.split(cmdUri)
  scriptUri = os.path.join(head, scriptName)

  scriptContent = ""
  with open(scriptUri,'r') as script:
    scriptContent = script.read()

  return scriptContent

def createProperties(comp):
  return

def createNode(parentNode):
  node_name = parentNode.Name + " " + GRASP_NODE_NAME
  node = parentNode.findNode(node_name)
  if not node:
    node = parentNode.createNode(VC_NODE, node_name)
  bound = parentNode.BoundDiagonal
  if mode.Value != "Collision Trigger":
    node.Offset = "Pos({}, {}, {})".format( 0, 0, bound.Z*2 )
  print "Set the desired grasping location for the", node.Name
  return node

def createFeatures(node):
  frame_name = node.Name + " TCP_Frame"
  frame = node.getFeature(frame_name)
  if frame is None:
    node.RootFeature.createFeature(VC_FRAME, frame_name)
    node.rebuild()

def createBehaviors(node):
  container_name = node.Name + " Container__HIDE__"
  container = node.findBehaviour(container_name)
  if not container:
    container = node.createBehaviour(VC_COMPONENTCONTAINER, container_name)
    tcp_frame = node.findFeature(node.Name + " TCP_Frame")
    container.Location = tcp_frame

  signal_name = node.Name + " Signal"
  signal = node.findBehaviour(signal_name)
  if not signal:
    signal = node.createBehaviour(VC_BOOLEANSIGNAL, signal_name)
  
  part_signal_name = node.Name + " ProductPresent"
  part_signal = node.findBehaviour(part_signal_name)
  if not part_signal:
    part_signal = node.createBehaviour(VC_BOOLEANSIGNAL, part_signal_name)

  grasp_script_name = node.Name + " Script"
  grasp_script = node.findBehaviour(grasp_script_name)
  if not grasp_script:
    grasp_script = node.createBehaviour(VC_SCRIPT, grasp_script_name)
    signal.Connections = [grasp_script]
    grasp_script.Script = getScript("grasp_script.py")
  
  trigger_script_name = node.Name + " TriggerScript"
  trigger_script = node.findBehaviour(trigger_script_name)
  script_file = None
  if mode.Value == "Servo Trigger":
    script_file = "servo_trigger_script.py"
  elif mode.Value == "Collision Trigger":
    script_file = "detection_trigger_script.py"
  if not trigger_script and script_file is not None:
    trigger_script = node.createBehaviour(VC_SCRIPT, trigger_script_name)
    default_script_content = getScript(script_file)
    trigger_script.Script = set_trigger_parameters(default_script_content)

#Hidas and shows the proeprties if servo trigger is added or not
def toggle_prop_visibility(arg):
  if arg.Value == "Default":
    servoProp.IsVisible = False
    triggerPointProp.IsVisible = False
    deltaProp.IsVisible = False
    det_interval.IsVisible = False
    delay_grasp.IsVisible = False
    delay_release.IsVisible = False
    rel_mode.IsVisible = False
    force_locs_p.IsVisible = False
  
  elif arg.Value == "Servo Trigger":
    servoProp.IsVisible = True
    triggerPointProp.IsVisible = True
    deltaProp.IsVisible = True
    det_interval.IsVisible = False
    delay_grasp.IsVisible = False
    delay_release.IsVisible = False
    rel_mode.IsVisible = False
    force_locs_p.IsVisible = False

  elif arg.Value == "Collision Trigger":
    det_interval.IsVisible = True
    delay_grasp.IsVisible = True
    delay_release.IsVisible = True
    servoProp.IsVisible = True
    rel_mode.IsVisible = True
    deltaProp.IsVisible = True
    force_locs_p.IsVisible = False
    triggerPointProp.IsVisible = False

#Clamp a int/flaot between two numbers
def clamp(n, min, max):
  if n < min:
      return min
  elif n > max:
      return max
  else:
      return n

def warning(s):
  print "Wizard '{}' error: {}".format(cmd.Name, s)


mode.OnChanged = toggle_prop_visibility
#force_locs_p.OnChanged = force_prod_locs
addState(first_state)
