from vcScript import *
import vcVector

comp = getComponent()
app = getApplication()
sim = getSimulation()
detection_node = getNode()
normal_vec = vcVector.new(0,0,1)
default_detection_volume = vcVector.new(50,50,50)
EPSILON = 10e-5

grasped_components = []

#Main loop. Listen to the grasping signal and trigger the corresponding action
def OnRun():
  global grasped_components
  DetectionVolumeVisiblity.Value = False
  grasped_components = []
  actions = ("Grasp", "Release")
  part_present_signal.signal(False)
  while True:
    triggerCondition(lambda: getTrigger() == grasp_signal)
    if grasp_signal.Value:
      ret = Grasp()
      if force_locs.Value and len(grasped_components) > 0:
        force_grasp_loc()
      action = actions[0]
    else:
      ret = Release()
      action = actions[1]
    if Verbose.Value:
      print "%s: Action \'%s\' executed (%s) " % (comp.Name, action, ret)


def force_grasp_loc():
  '''Forces an offset for the part when it is grasped'''
  #Check which axis should be forced
  force_x, force_y, force_z = enabled_force_offsets()
  if not any([force_x, force_y, force_z]):
    return
  
  grasping_node.update()
  part = grasped_components[0]
  product_mtx = part.PositionMatrix
  new_x = product_mtx.P.X
  new_y = product_mtx.P.Y
  new_z = product_mtx.P.Z
  
  bound_ctr = part.BoundCenter
  bound_diag = part.BoundDiagonal
  if force_x:
    new_x = force_pick.Value.X
  if force_y:
    new_y = force_pick.Value.Y
  if force_z: #Z uses offset from top of bounding box
    product_top_offset = bound_ctr.Z + bound_diag.Z
    new_z = product_top_offset+force_pick.Value.Z
  product_mtx.P = vcVector.new(new_x, new_y, new_z)
  part.PositionMatrix = product_mtx


def force_release_loc(part):
  '''Forces a Z offset for the part when it is released'''
  #Check which axis should be forced
  force_x, force_y, force_z = enabled_force_offsets()
  if not any([force_x, force_y, force_z]):
    return
  
  parent = part.Parent
  if not parent:
    print "No parent component found for releasing"
    return
  product_mtx = part.PositionMatrix
  new_x = product_mtx.P.X
  new_y = product_mtx.P.Y
  new_z = product_mtx.P.Z
  bound_ctr = parent.BoundCenter
  bound_diag = parent.BoundDiagonal
  if force_x:
    new_x = force_place.Value.X
  if force_y:
    new_y = force_place.Value.Y
  if force_z: #Z uses offset from top of bounding box
    parent_top_offset = bound_ctr.Z + bound_diag.Z
    new_z = parent_top_offset + force_place.Value.Z
  product_mtx.P = vcVector.new(new_x, new_y, new_z)
  part.PositionMatrix = product_mtx
  part.update()


#Check if any offsets should be forced
def enabled_force_offsets():
  force_x, force_y, force_z = False, False, False
  for sub_string in ["true", "all", "yes"]:
    if sub_string in force_locs.Value.lower():
      force_x, force_y, force_z = True, True, True
  if "x" in force_locs.Value.lower():
    force_x = True
  if "y" in force_locs.Value.lower():
    force_y = True
  if "z" in force_locs.Value.lower():
    force_z = True
  return force_x, force_y, force_z

  
def Grasp():
  global grasped_components
  if MultiGrasp.Value:
    count = 0
    multi_grasp_list = []
    while True:
      part = DetectionHelper(DetectionVolume.Value)
      if not part:
        break
      multi_grasp_list.append(part)
      root_phys_entities = part.getBehavioursByType('rPhysicsEntity')
      for root_phys_entity in root_phys_entities:
        physicsType = root_phys_entity.getProperty('Physics Type')
        if physicsType.Value == 0:
          physicsType.Value = 2 # switch to kinematic
      # assembly handling
      if GraspIncludeEmptyAssemblies.Value == "No -Delete": 
        grasp_container.grab(part, assemblymode=VC_GRAB_ASSEMBLY_DELETE_IF_EMPTY)
      elif GraspIncludeEmptyAssemblies.Value == "Yes": 
        grasp_container.grab(part, assemblymode=VC_GRAB_ASSEMBLY_GRAB)
      else: grasp_container.grab(part, assemblymode=VC_GRAB_ASSEMBLY_KEEP)
      grasped_components.append( part )
      count += 1
    if count > 0:
      #grasped_components_name = grasped_components[-1].Name
      grasped_components_name = [x.Name for x in grasped_components]
      part_present_signal.signal(True)
      return 'Ok. {} component(s): "{}", Grasped.'.format(count,''.join(c for c in str(grasped_components_name) if c not in '[]' and c))
    else:
      part_present_signal.signal(False)
      return "No component detected to grasp"
  else:
    part = DetectionHelper(DetectionVolume.Value)
    if part:
      root_phys_entities = part.getBehavioursByType('rPhysicsEntity')
      for root_phys_entity in root_phys_entities:
        physicsType = root_phys_entity.getProperty('Physics Type')
        if physicsType.Value == 0:
          physicsType.Value = 2 # switch to kinematic

      # assembly handling
      if GraspIncludeEmptyAssemblies.Value == "No -Delete": 
        grasp_container.grab(part, assemblymode=VC_GRAB_ASSEMBLY_DELETE_IF_EMPTY)
      elif GraspIncludeEmptyAssemblies.Value == "Yes": 
        grasp_container.grab(part, assemblymode=VC_GRAB_ASSEMBLY_GRAB)
      else: grasp_container.grab(part, assemblymode=VC_GRAB_ASSEMBLY_KEEP)
      
      grasped_components.append( part )
      #grasped_components_name = [x.Name for x in grasped_components]
      grasped_components_name = grasped_components[-1].Name
      part_present_signal.signal(True)
      return 'Ok. Grasped component: "{}"'.format(''.join(c for c in str(grasped_components_name) if c not in '[]'))
    else:
      part_present_signal.signal(False)
      return "No component detected to grasp"


#Check if the release_node is a list. If so, check if any element matches the release filters and return that element
def custom_release_function(release_node):
  if isinstance(release_node, list):
    release_candidates = release_node
    release_node = release_candidates[0]
    filter_names = [x.strip() for x in ReleaseNameFilter.Value.split(',')]
    for candidate in release_candidates:
      for filter_name in filter_names:
        if filter_name in candidate.Name:
          release_node = candidate
          break
  return release_node


#
def Release():
  global grasped_components
  # Find first the part to be released
  set_phys = False
  part = None
  
  try:
    if len(grasped_components) == 0:
      return "No component to release"
    elif grasped_components[0] is None or grasped_components[0].Container != grasp_container:
      grasped_components.pop(0)
      return "Grasped component was removed"
  except:
    grasped_components = []
    message = "Part was deleted by user."
    return message
 
  if len(grasped_components) == 0:
    return "No component to release"
  elif grasped_components[0] is None or grasped_components[0].Container != grasp_container:
    grasped_components.pop(0)
    return "Grasped component was removed"
  release_list = []
  if MultiGrasp.Value:
    release_list = grasped_components
  else:
    release_list.append(grasped_components[0])
  count_to_physics = 0
  count_to_node = 0
  count_to_world = 0
  count_to_container = 0
  count_to_world_container = 0
  first_part = None
  release_list_name = [x.Name for x in release_list]
  for part in release_list:
    grasped_components.remove( part )
    if ReleaseToPhysics.Value == 'Always':
      set_phys = True
      release_node = sim.World
    elif first_part and BundleReleasedComponents.Value:
      release_node = first_part
    elif ReleaseToWorld.Value:
      release_node = sim.World
      if ReleaseToPhysics.Value != 'Never':
        set_phys = True
    else:
      release_node = ReleaseHelper( part, release_detection_tol_prop.Value )
      if not release_node:
        if ReleaseToPhysics.Value != 'Never':
          set_phys = True
        release_node = sim.World
    
    #Check if the release_node is a list. If so, check if any element matches the release filters and return that element
    release_node = custom_release_function(release_node)

    #print "release_node: ",release_node
    #release_node = search_product_by_name(release_node,ReleaseNameFilter.Value)
    
    # Calculate position in relative to the detected component (NOTE: component origin)
    pos_m = release_node.InverseWorldPositionMatrix * part.WorldPositionMatrix
    pos = pos_m.P
    # Extract all containers what this node owns (containers where to release)
    if release_node != sim.World:
      ContainerList = release_node.getBehavioursByType(VC_CONTAINER)
    else:
      ContainerList = []
    if first_part and BundleReleasedComponents.Value:
      # Do not use container for bundled components
      ContainerList = []
    # Search closest container in the detected component
    BestContainer = None
    BestProximity = 1e10
    for cont in ContainerList:
      dist = cont.getContainerProximity( pos )
      if dist < 0.0: # Container with no physical limits return -1.0 value
        dist = 1e5
      if dist < BestProximity:
        BestContainer = cont
        BestProximity = dist
    # Release component to the closest container
    if BestContainer:
      # print "Container \'%s\' found at proximity = %f" % (BestContainer.Name, BestProximity)
      # assembly handling
      if GraspIncludeEmptyAssemblies.Value == "No -Delete": BestContainer.grab(part, assemblymode=VC_GRAB_ASSEMBLY_DELETE_IF_EMPTY)
      elif GraspIncludeEmptyAssemblies.Value == "Yes": BestContainer.grab(part, assemblymode=VC_GRAB_ASSEMBLY_GRAB)
      else: BestContainer.grab(part, assemblymode=VC_GRAB_ASSEMBLY_KEEP)
      #
      if force_locs.Value and part:
        force_release_loc(part)
      
      message = ""
      if release_node.Parent:
        count_to_container += 1
        message = "Ok, Released component {} to Node: {} container".format(''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Parent.Name)
      else:
        count_to_world_container += 1
        message = "Ok, Released component {} to world container".format(''.join(c for c in str(release_list_name) if c not in '[]'))
      if not MultiGrasp.Value:
        if set_phys:
          message = "Ok, Released component {} to physics".format(''.join(c for c in str(release_list_name) if c not in '[]'))
          setPhysics(part)
    else:
      # print "No container found to release component, just connect it hierarchically"
      pos_m = release_node.InverseWorldPositionMatrix * part.WorldPositionMatrix
      part.PositionMatrix = pos_m
      release_node.attach(part, False)
      #
      if force_locs.Value and part:
        force_release_loc(part)
      message = "Ok, "
      if release_node.Parent:
        count_to_node += 1
        if release_node.Parent.Name == 'ROOT':
          message += 'Attached component {} directly to Node: ROOT in Component: "{}".'\
                    .format(''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Component.Name)
        elif release_node.Component.Visible == False:
          message += 'Attached component {0} directly to Node: "{1}" in Component: "{2}".Attached COMPONENT: "{2}" is IN-VISIBLE'\
                     .format(''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name, release_node.Component.Name)
        elif release_node.Visible == False:
          message += 'Attached component {0} directly to Node: "{1}" in Component: "{2}".Attached NODE: "{1}" is IN-VISIBLE'\
                       .format(''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name, release_node.Component.Name)
        else:
          invisible_node = CheckNodeParentVisiblity(release_node)
          if not invisible_node:
            message += 'Attached component {} directly to Node: "{}" in Component: "{}".'\
                       .format(''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name, release_node.Component.Name)
          else:
            message += 'Attached component {} directly to Node: "{}" in Component: "{}".Attached NODE: "{}" is IN-VISIBLE'\
                       .format(''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name, release_node.Component.Name, invisible_node.Name)
      else:
        count_to_world += 1
        message = "Ok, Attached directly to world"
      if not MultiGrasp.Value:
        if set_phys:
          message = "Ok, Released to physics"
          setPhysics(part)
        part_present_signal.signal(False)
        return message
    if set_phys:
      count_to_physics += 1
      setPhysics(part)
    if not first_part and BundleReleasedComponents.Value:
      first_part = part
      first_part.update()
  #endfor
  message = "Ok, "
  if count_to_physics:
    message += "Released %i to physics. " % count_to_physics
  else:
    if count_to_node == 1:
      message += "Attached directly to node, "
    elif count_to_node > 1:
      if release_node.Parent.Name == 'ROOT':
        message += '{} component(s): {}, Attached directly to Node: ROOT in Component: "{}".'\
                    .format(count_to_node, ''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Component.Name)
      elif release_node.Component.Visible == False:
        message += '{0} component(s): {1},  Attached directly to Node: "{2}" in Component: "{3}".Attached COMPONENT: "{3}" is IN-VISIBLE'\
                  .format(count_to_node,''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name, release_node.Component.Name)
      elif release_node.Visible == False:
        message += '{0} component(s): {1}, Attached directly to Node: "{2}" in Component: "{3}".Attached NODE: "{2}" is IN-VISIBLE'\
                    .format(count_to_node, ''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name, release_node.Component.Name)
      else:
        invisible_node = CheckNodeParentVisiblity(release_node)
        if not invisible_node:
          message += '{} component(s): {}, Attached directly to Node: "{}" in Component: "{}".'\
                    .format(count_to_node, ''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name, release_node.Component.Name)
        else:
          message += '{} component(s): {}, Attached directly to Node: "{}" in Component: "{}".Attached NODE: "{}" is IN-VISIBLE'\
                    .format(count_to_node, ''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name, release_node.Component.Name, invisible_node.Name)

    if count_to_world == 1:
      message += "Attached component: {} directly to world.".format(''.join(c for c in str(release_list_name) if c not in '[]'))
    elif count_to_world > 1:
          message += " {} component(s): {}, Attached directly to world".format(count_to_world, ''.join(c for c in str(release_list_name) if c not in '[]'))
    if count_to_container == 1:
      message += 'Released component: {} to Node container in Component: {}'.format(''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name)
    elif count_to_container > 1:
          message += '{} component(s): {}, Released to Node in Component: "{}".'\
                    .format(count_to_container, ''.join(c for c in str(release_list_name) if c not in '[]'), release_node.Name)
    if count_to_world_container == 1:
      message += "Released component: {} to world container"\
                  .format(''.join(c for c in str(release_list_name) if c not in '[]'))
    elif count_to_world_container > 1:
          message += "{} component(s): {}, Released to world container."\
                      .format(count_to_world_container, ''.join(c for c in str(release_list_name) if c not in '[]'))
  part_present_signal.signal(False)
  return message

#Tries to check the grasp filter and collision detection to return only wanted grasp candidate
def custom_grasp_function(GraspNameFilter, coll_detector):
  filter_names = [x.strip() for x in GraspNameFilter.Value.split(',')]
  part_candidates = []
  if coll_detector is not None:
    i = 0
    while i < coll_detector.HitCount:
      part_test = coll_detector.getHitNodeA(i)
      if part_test is None:
        break
      part_candidates.append(part_test.Component)
      i += 1
  for candidate in part_candidates:
    for filter_name in filter_names:
      if filter_name in candidate.Name:
        return candidate

#
def DetectionHelper(bound):
  global grasp_container
  # Define bound corner points when original bound vector is thought as bound box diagonal vector
  v1 = bound*(-0.5)
  v2 = bound*0.5

  # Following code sets a specific tool and extracts tool matrix in world coordinate system
  # This transformation is then applied to the bound box defining volume.
  sim.update()
  tnode = detection_node
  tmatrix = detection_frame.PositionMatrix
  m = tnode.WorldPositionMatrix * tmatrix
  # a) Create volume bbox detector and test if geometry bbox is fully inside the detection volume
  part = None
  if GraspNameFilter.Value == "":
    vol_detector = sim.newVolumeDetector()
    vol_detector.TestMethod = VC_INSIDE
    vol_detector.Corner1 = v1
    vol_detector.Corner2 = v2
    vol_detector.Transformation = m
    node_list = get_grasp_candidates()
    vol_detector.NodeList = node_list
    hit = vol_detector.testOneCollision()
    if hit:
      part = vol_detector.getHitNode(0).Component
      part = prevent_grasp_filter(part)

  # b) If no geometry bbox was fully inside the detection volume, create new colision detector
  # using volume as detection geometry
  coll_detector = None
  if not part:
    coll_detector = sim.newCollisionDetector()
    node_list = get_grasp_candidates()
    coll_detector.NodeListA = node_list
    if GraspNameFilter.Value == "":
      hit = coll_detector.testOneBBoxCollision(v1,v2,m)
    else:
      hit = coll_detector.testAllBBoxCollisions(v1,v2,m)
    if hit:
      part = coll_detector.getHitNodeA(0).Component
      part = prevent_grasp_filter(part)
    if not part:
      return None
      
  if GraspNameFilter.Value != "":
    candidate = custom_grasp_function(GraspNameFilter, coll_detector)
    if candidate is not None:
      return candidate
    #part = search_product_by_name(part,filter_name)
  part = prevent_grasp_filter(part)
  return part


def prevent_grasp_filter(part):
  '''Check if prevent grap filter is set. Return none if a substring from the filter matches with given part name. Otherwise return part'''
  if prevent_grasp_filter_prop.Value != "" and part is not None:
    if prevent_grasp_filter_prop.Value.lower() in part.Name.lower():
      return None
  return part

def search_product_by_name(temp_part,filter_name):
  #print "source part: ",part.Name
  if filter_name in temp_part.Name:
    return temp_part
  # walk childs
  children = temp_part.ComponentChildren
  while len(children) > 0:
    #print "children: ",children
    child = children.pop(0)
    #print "child: ",child
    if filter_name in child.Name:
      #print "Grasp child: ",child.Name
      return child
    children.extend(child.ComponentChildren)
  # walk parents
  parents = [temp_part.Parent]
  while len(parents) > 0:
    parent = parents.pop(0)
    if parent == None or parent.Name == "ROOT":
      break
    if filter_name in parent.Name:
      return parent
    parents.extend([parent.Parent])
  return temp_part

#
def get_grasp_candidates():
  grasp_candidates = [ (sim.World, VC_NODELIST_INCLUDE, VC_NODELIST_TREE ) , (comp, VC_NODELIST_EXCLUDE, VC_NODELIST_COMPONENT )] #VC_NODELIST_TREE
  grasp_candidates.extend([(x, VC_NODELIST_EXCLUDE, VC_NODELIST_COMPONENT ) for x in app.Components if exclude_grasp_object_test(x)])
  if GraspOnlyProducts.Value:
    grasp_candidates.extend([(x, VC_NODELIST_EXCLUDE, VC_NODELIST_COMPONENT ) for x in app.Components if x.Product == None])
  return grasp_candidates

#
def exclude_grasp_object_test(icomponent):
  if icomponent in ExcludeGraspingComponents.Value:
    return True
  if not DetectHiddenComponents.Value and not icomponent.Visible:
    return True
  if not DetectLockedComponents.Value and icomponent.Locked:
    return True
  for ignore_prop_name in ignore_as_grasp_object_filter_props:
    ignore_prop = icomponent.getProperty(ignore_prop_name)
    if ignore_prop and ignore_prop.Value:
      return True

#
def ReleaseHelper( part, detection_tolerance ):
  global robot
  part.update()
  # Detect where to release using offset at "gravity" direction
  # Basically just translate detected component in the world z axis direction downwards with tolerance value
  gravity = vcVector.new(0,0,-detection_tolerance)
  # Calculate gravity at node parent coordinate system ( reset translate component and use just rotation)
  invert_coord = part.Parent.InverseWorldPositionMatrix
  invert_coord.P = vcVector.new()
  node_gravity = invert_coord * gravity
  # Translate node using gravity
  # print "Node gravity X=%f Y=%f Z=%f" % (node_gravity.X,node_gravity.Y,node_gravity.Z)
  orig_m = part.PositionMatrix
  pos_m = part.PositionMatrix
  pos = pos_m.P
  pos += node_gravity
  pos_m.P = pos
  part.PositionMatrix = pos_m
  if UpdateWorldIOnRelease.Value:
    sim.update()
  else:
    part.update()
  release_node = None
  detector = sim.newCollisionDetector()
  detector.NodeListA = [ (part, VC_NODELIST_INCLUDE, VC_NODELIST_COMPONENT ) ]
  detector.NodeListB = get_target_candidates()
  if ReleaseNameFilter.Value == "":
    hit = detector.testOneCollision()
  else:
    hit = detector.testAllCollisions()
  # Restore location (minus gravity offset)
  part.PositionMatrix = orig_m
  part.update()
  if hit:
    if ReleaseNameFilter.Value == "":
      return detector.getHitNodeB(0)
    else:
      release_candidates = []
      i = 0
      while i < detector.HitCount:
        test_node = detector.getHitNodeB(i)
        if test_node is None:
          break
        release_candidates.append(test_node)
        i += 1
      return release_candidates
  else:
    return None


def setPhysics(part):
  root_phys_entities = part.getBehavioursByType('rPhysicsEntity')
  if root_phys_entities:
    root_phys_entity = root_phys_entities[0]
  else:
    root_phys_entity = part.createBehaviour('rPhysicsEntity', 'PhysicsEntity')
  physicsType = root_phys_entity.getProperty('Physics Type')
  physicsType.Value = 0


def updateDetectionLocProp(*args):
  detect_frameloc_prop.Value = detection_node.PositionMatrix.P
  if DetectionVolumeVisiblity.Value:
    GenerateDetectionGeometrySet()


def updateDetectionLoc(*args):
  pos = detect_frameloc_prop.Value
  detection_node.Offset = "Tx({}).Ty({}).Tz({})".format( pos.X, pos.Y, pos.Z )
  if DetectionVolumeVisiblity.Value:
    GenerateDetectionGeometrySet() #TODO check that the visualization moves when frame is moved
  comp.rebuild()


def get_target_candidates():
  target_candidates = [ (sim.World, VC_NODELIST_INCLUDE, VC_NODELIST_TREE ) , (detection_node.Parent, VC_NODELIST_EXCLUDE, VC_NODELIST_TREE )] #(comp, VC_NODELIST_EXCLUDE, VC_NODELIST_TREE )
  target_candidates.extend([(x, VC_NODELIST_EXCLUDE, VC_NODELIST_COMPONENT ) for x in app.Components if exclude_release_target_test(x)]) #VC_NODELIST_COMPONENT
  return target_candidates


def exclude_release_target_test(icomponent):
  if icomponent in ExcludeReleasingToComponents.Value:
    return True
  if not DetectHiddenComponents.Value and not icomponent.Visible:
    return True
  for ignore_prop_name in ignore_as_release_target_filter_props:
    ignore_prop = icomponent.getProperty(ignore_prop_name)
    if ignore_prop and ignore_prop.Value:
      return True


def CheckNodeParentVisiblity(node): 
  if node.Parent.Name != "ROOT":
    if node.Visible == False:
      return node
    else:
      node = node.Parent 
      return CheckNodeParentVisiblity(node)


ignore_as_release_target_filter_props = ['ExcludeRobotReleasingTo']
ignore_as_grasp_object_filter_props = ['ExcludeRobotGrasping']

def DetectionVolumeChange(prop):
  if DetectionVolumeVisiblity.Value:
    GenerateDetectionGeometrySet()


def GenerateDetectionGeometrySet():
  current_detection_volume = DetectionVolume
  box = current_detection_volume.Value * 0.5 
  tool_node = detection_node
  tool_matrix = detection_frame.PositionMatrix
  tool_frame_matrix = comp.InverseWorldPositionMatrix * tool_node.WorldPositionMatrix * tool_matrix

  geo = comp.UserGeometry
  geo.clear()

  p1 = tool_frame_matrix * vcVector.new(box.X, box.Y, box.Z)
  p2 = tool_frame_matrix * vcVector.new(box.X, -box.Y, box.Z)
  p3 = tool_frame_matrix * vcVector.new(-box.X, box.Y, box.Z)
  p4 = tool_frame_matrix * vcVector.new(-box.X, -box.Y, box.Z)
  p5 = tool_frame_matrix * vcVector.new(box.X, box.Y, -box.Z)
  p6 = tool_frame_matrix * vcVector.new(box.X, -box.Y, -box.Z)
  p7 = tool_frame_matrix * vcVector.new(-box.X, box.Y, -box.Z)
  p8 = tool_frame_matrix * vcVector.new(-box.X, -box.Y, -box.Z)

  color = app.findMaterial('transp_red')
  tset = geo.createGeometrySet(VC_TRIANGLESET) 
  tset.UseBackface = True
  tset.Material = color

  tp1 = tset.addPoint(p1)
  tp2 = tset.addPoint(p2)
  tp3 = tset.addPoint(p3)
  tp4 = tset.addPoint(p4)
  tp5 = tset.addPoint(p5)
  tp6 = tset.addPoint(p6)
  tp7 = tset.addPoint(p7)
  tp8 = tset.addPoint(p8)

  tset.addTriangle(tp1,tp2,tp4)    #tri_Right1
  tset.addTriangle(tp4,tp3,tp1) #tri_Right2
  tset.addTriangle(tp1,tp3,tp7) #tri_Top1
  tset.addTriangle(tp7,tp5,tp1) #tri_Top2
  tset.addTriangle(tp6,tp8,tp7)   #tri_Left1
  tset.addTriangle(tp7,tp5,tp6)   #tri_Left2 
  tset.addTriangle(tp6,tp2,tp4) #tri_Bottom1
  tset.addTriangle(tp4,tp8,tp6) #tri_Bottom2
  tset.addTriangle(tp8,tp4,tp3)   #tri_Back1
  tset.addTriangle(tp3,tp7,tp8)   #tri_Back2 
  tset.addTriangle(tp6,tp2,tp1)  #tri_Front1 
  tset.addTriangle(tp1,tp5,tp6)  #tri_Front2 
  tset.update()

  app.render()


def ChangeDetectionVisiblity(prop):
  if prop.Value:
    DetectionVolume.IsVisible = True
    GenerateDetectionGeometrySet()  
  else:
    comp.UserGeometry.clear()
    DetectionVolume.IsVisible = False  


def confirmProperty(name,defaultvalue,prop_type=VC_REAL,constraints=VC_PROPERTY_DEFAULT):
     global comp
     prop = comp.getProperty(name)
     if prop_type == VC_VECTOR:
        if prop == None:
          prop = comp.createProperty(prop_type,name,constraints)
          if defaultvalue:
            if type(defaultvalue) == type(vcVector.new()):
              prop.Value = defaultvalue
              return prop
          prop.Value = vcVector.new()

     else:
        if prop == None:
         prop = comp.createProperty(prop_type,name,constraints)
         prop.Value = defaultvalue
     return prop


# BEHAVIORS AND FEATURES
grasp_container = detection_node.findBehaviour(detection_node.Name + " Container__HIDE__")
grasp_signal = detection_node.findBehaviour(detection_node.Name + " Signal")
detection_frame = detection_node.findFeature(detection_node.Name + " TCP_Frame")
part_present_signal = detection_node.findBehaviour(detection_node.Name + " ProductPresent")


# PROPERTIES
detect_frameloc_prop = confirmProperty(detection_node.Name + "::TCPFrameLocation",vcVector.new(0,0,0),VC_VECTOR)
release_detection_tol_prop = confirmProperty(detection_node.Name + "::GravityDirection",25.0,VC_REAL)
Verbose = confirmProperty(detection_node.Name + "::DisplayMessages",False,VC_BOOLEAN)
ReleaseToWorld = confirmProperty(detection_node.Name + "::ReleaseToWorld",False,VC_BOOLEAN)
ReleaseToPhysics = confirmProperty(detection_node.Name + "::ReleaseToPhysics",'Never',VC_STRING, VC_PROPERTY_STEP)
ReleaseToPhysics.StepValues = ['Never', 'If Target not Detected', 'Always']
MultiGrasp = confirmProperty(detection_node.Name + "::MultiGrasp",False,VC_BOOLEAN)
BundleReleasedComponents = confirmProperty(detection_node.Name + "::BundleReleasedComponents",False,VC_BOOLEAN)
UpdateWorldIOnRelease = confirmProperty(detection_node.Name + "::UpdateWorldOnRelease",False,VC_BOOLEAN)
DetectLockedComponents = confirmProperty(detection_node.Name + "::DetectLockedComponents",False,VC_BOOLEAN)
DetectHiddenComponents = confirmProperty(detection_node.Name + "::DetectHiddenComponents",False,VC_BOOLEAN)
GraspOnlyProducts = confirmProperty(detection_node.Name + "::GraspOnlyProducts",False,VC_BOOLEAN)
GraspNameFilter = confirmProperty(detection_node.Name + "::GraspNameFilter","",VC_STRING)
ReleaseNameFilter = confirmProperty(detection_node.Name + "::ReleaseNameFilter","",VC_STRING)
prevent_grasp_filter_prop = confirmProperty(detection_node.Name + "::PreventGraspFilter","",VC_STRING)
ExcludeReleasingToComponents = confirmProperty(detection_node.Name + "::ExcludeReleasingTo",[],"List<Ref<Component> >")
ExcludeReleasingToComponents.IsVisible = True
ExcludeGraspingComponents = confirmProperty(detection_node.Name + "::ExcludeGrasping",[],"List<Ref<Component> >")
ExcludeGraspingComponents.IsVisible = True

'''
Detection volume visualization
'''
DetectionVolumeVisiblity = confirmProperty(detection_node.Name + "::ShowDetectionVolume",False,VC_BOOLEAN)
DetectionVolumeVisiblity.ChangeOnRebuild = True
DetectionVolumeVisiblity.Value = False
DetectionVolumeVisiblity.OnChanged = ChangeDetectionVisiblity

DetectionVolume = confirmProperty(detection_node.Name + "::DetectionVolumeSize",default_detection_volume,VC_VECTOR)
DetectionVolume.Quantity =  "Distance"
DetectionVolume.IsVisible = False
DetectionVolume.OnChanged = DetectionVolumeChange

'''
Assembly behaviour
'''
GraspIncludeEmptyAssemblies = confirmProperty(detection_node.Name + "::GraspIncludeEmptyAssemblies",'No -Delete',VC_STRING, VC_PROPERTY_STEP)
GraspIncludeEmptyAssemblies.StepValues = ['No -Delete', 'No -Leave', 'Yes']

'''
Collision detection mode
'''
grasping_node = getNode()
force_locs = confirmProperty(grasping_node.Name + "::ForceProductOffset", "", VC_STRING)
force_pick = confirmProperty(grasping_node.Name + "::PickOffset", vcVector.new(), VC_VECTOR)
force_place = confirmProperty(grasping_node.Name + "::PlaceOffset", vcVector.new(), VC_VECTOR)


# group_index = 0
# Verbose.Group = group_index = group_index+10
# ReleaseToWorld.Group = group_index = group_index+10
# ReleaseToPhysics.Group = group_index = group_index+10
# MultiGrasp.Group = group_index = group_index+10

detection_node.OnNodeTransform = updateDetectionLocProp
detect_frameloc_prop.OnChanged = updateDetectionLoc
updateDetectionLocProp()

# #Not used now due to performance impact
# def DetectionVolumeOnRender(simulation):
#   if simulation.IsRunning and DetectionVolumeVisiblity.Value:
#     DetectionVolumeVisiblity.Value = False
#   else:
#     if DetectionVolumeVisiblity.Value == True:
#       GenerateDetectionGeometrySet()


#app.OnRender = DetectionVolumeOnRender
